@@ -1,6 +1,8 @@
 # -*- coding: utf-8 -*-
 """The Mosaic Python API
 
+.. moduleauthor:: Konrad Hinsen
+
 This module provides abstract base classes that define the Mosaic
 Python API and implement validation code. They also provide a few
 convenience functions implemented in terms of the raw API.
@@ -10,12 +12,20 @@ implement all the abstract properties.
 
 """
 
+#-----------------------------------------------------------------------------
+#       Copyright (C) 2013 The Mosaic Development Team
+#
+#  Distributed under the terms of the BSD License.  The full license is in
+#  the file LICENSE.txt, distributed as part of this software.
+#-----------------------------------------------------------------------------
+
 from abc import ABCMeta, abstractproperty
 import collections
 import itertools as IT
 import re
 
 import numpy as N
+import numpy.linalg as LA
 
 import mosaic.utility
 
@@ -25,7 +35,7 @@ import mosaic.utility
 # versions. An increase in the major number indicated an incompatibility
 # with preceding versions.
 
-MOSAIC_VERSION = (0, 1)
+MOSAIC_VERSION = (1, 0)
 
 
 # Base class for all classes that represent top-level data items,
@@ -81,18 +91,6 @@ class MosaicAtom(object):
         """
         raise NotImplementedError()
 
-    @abstractproperty
-    def parent(self):
-        "The fragment to which the atom belongs."
-        raise NotImplementedError()
-
-    @abstractproperty
-    def unique_id(self):
-        """An object that uniquely identifies this atom, to be used
-        in identity tests (is, is not, ...).
-        """
-        raise NotImplementedError()
-
     @abstractproperty
     def name(self):
         """An ASCII string describing the type of the atom. For 'real'
@@ -200,8 +198,6 @@ class MosaicAtom(object):
         :raises ValueError: if the object is not valid Mosaic data
         """
         validate_label(self.label, "Atom.label")
-        if self.parent is not None:
-            validate_type(self.parent, MosaicFragment, "Fragment.parent")
         validate_value(self.type, self._allowed_types, "Atom type")
         validate_label(self.name, "Atom.name")
         if self.type == 'element':
@@ -271,20 +267,6 @@ class MosaicFragment(collections.Mapping):
         """
         raise NotImplementedError()
 
-    @abstractproperty
-    def parent(self):
-        """The parent fragment to which the fragment belongs.
-        None for top-level fragments.
-        """
-        raise NotImplementedError()
-
-    @abstractproperty
-    def unique_id(self):
-        """An object that uniquely identifies this fragment, to be used
-        in identity tests (is, is not, ...).
-        """
-        raise NotImplementedError()
-
     @abstractproperty
     def fragments(self):
         """Sequence of sub-fragments, may be empty.
@@ -419,8 +401,6 @@ class MosaicFragment(collections.Mapping):
         """
         validate_label(self.label, "Fragment.label")
         validate_label(self.species, "Fragment.species")
-        if self.parent is not None:
-            validate_type(self.parent, MosaicFragment, "Fragment.parent")
         validate_value(self.is_polymer, [True, False], "Fragment.is_polymer")
         if self.is_polymer:
             validate_value(self.polymer_type, self._polymer_types,
@@ -429,16 +409,12 @@ class MosaicFragment(collections.Mapping):
         labels = set()
         for f in self.fragments:
             f.validate()
-            if f.parent.unique_id != self.unique_id:
-                raise ValueError("Invalid fragment tree structure")
             if f.label in labels:
                 raise ValueError("Label %s occurs more than once" % f.label)
             labels.add(f.label)
         validate_sequence(self.atoms, MosaicAtom, "Fragment.atoms")
         for a in self.atoms:
             a.validate()
-            if a.parent.unique_id != self.unique_id:
-                raise ValueError("Invalid fragment tree structure")
             if a.label in labels:
                 raise ValueError("Label %s occurs more than once" % a.label)
             labels.add(a.label)
@@ -461,13 +437,52 @@ class MosaicFragment(collections.Mapping):
                                                for f in self.fragments),
                         iter(self.atoms))
 
+    def recursive_atom_path_iterator(self):
+        """An iterator over the atom paths in the fragment, including the
+        atoms in sub-fragments.
+        """
+        for f in self.fragments:
+            l = f.label
+            for ap in f.recursive_atom_path_iterator():
+                yield l + '.' + ap
+        for a in self.atoms:
+            yield a.label
+
     def recursive_bond_iterator(self):
         """An iterator over the bonds in the fragment, including the
         bonds in sub-fragments.
         """
-        return IT.chain(IT.chain.from_iterable(f.recursive_bond_iterator()
-                                               for f in self.fragments),
-                        iter(self.bonds))
+        for f in self.fragments:
+            l = f.label
+            for a1, a2, order in f.recursive_bond_iterator():
+                yield (l + '.' + a1, l + '.' + a2, order)
+        for b in self.bonds:
+            yield b
+
+    def site_to_atom_index_mapping(self):
+        """
+        :returns: an array whose element [s] is the atom index
+                  corresponding to site index s.
+        :rtype: numpy.ndarray
+        """
+        ns = [a.number_of_sites for a in self.recursive_atom_iterator()]
+        return N.repeat(N.arange(len(ns)), ns)
+
+    def _atom_to_site_index_mapping(self):
+        ns = [a.number_of_sites for a in self.recursive_atom_iterator()]
+        return N.add.accumulate(ns)
+
+    def atom_to_site_index_mapping(self):
+        """
+        :returns: an array whose element [s] is the site index
+                  corresponding to the first site of the atom with
+                  atom index s.
+        :rtype: numpy.ndarray
+        """
+        m = self._atom_to_site_index_mapping()
+        m[1:] = m[:-1]
+        m[0] = 0
+        return m
 
     # Mapping interface
 
@@ -650,10 +665,10 @@ class MosaicUniverse(MosaicDataItem):
                                 repr(other.convention)))
         for (r1, t1), (r2, t2) in zip(self.symmetry_transformations,
                                       other.symmetry_transformations):
-            if r1 != r2:
+            if (r1 != r2).any():
                 raise ValueError("rotation matrices differ: %s != %s"
                                  % (str(r1), str(r2)))
-            if t1 != t2:
+            if (t1 != t2).any():
                 raise ValueError("translation vectors differ: %s != %s"
                                  % (str(t1), str(t2)))
         for (sf, sc), (of, oc) in zip(self.molecules, other.molecules):
@@ -703,6 +718,10 @@ class MosaicUniverse(MosaicDataItem):
                             rot.dtype == N.float64
                             and trans.dtype == N.float64,
                             "rotation and translation must be float64")))
+        if self.cell_shape == "infinite" \
+           and len(self.symmetry_transformations) > 0:
+            raise ValueError("Symmetry transformations are allowed "
+                             "only in periodic universes")
         for f, n in self.molecules:
             f.validate()
             if not isinstance(n, int):
@@ -720,6 +739,118 @@ class MosaicUniverse(MosaicDataItem):
                 raise ValueError("Universe.%s is %s, should be %s"
                                  % (property, str(value), str(reference)))
 
+    # Methods based on API properties
+
+    def recursive_atom_iterator(self):
+        """An iterator over the atoms in the universe.
+        """
+        for fragment, count in self.molecules:
+            for _ in range(count):
+                for a in fragment.recursive_atom_iterator():
+                    yield a
+
+    def bond_index_array(self):
+        """Returns an integer array of shape (N, 2), where N
+        is the total number of bonds in the universe. The entries
+        [i, 0] and [i, 1] refer to the two atoms that are connected
+        by bond i. The entry [i, 0] is smaller than the entry [i, 1].
+
+        :returns: the bond index array
+        :rtype: numpy.ndarray
+        """
+        natoms = 0
+        bonds = []
+        for fragment, count in self.molecules:
+            f_paths = list(fragment.recursive_atom_path_iterator())
+            f_bonds = [sorted((f_paths.index(a1), f_paths.index(a2)))
+                       for a1, a2, order in fragment.recursive_bond_iterator()]
+            for _ in range(count):
+                bonds.extend([(natoms+a1, natoms+a2) for a1, a2 in f_bonds])
+                natoms += len(f_paths)
+        return N.array(bonds)
+
+    def site_to_atom_index_mapping(self):
+        """
+        :returns: an array whose element [s] is the atom index
+                  corresponding to site index s.
+        :rtype: numpy.ndarray
+        """
+        natoms = 0
+        total = []
+        for fragment, count in self.molecules:
+            per_fragment = fragment.site_to_atom_index_mapping()
+            f_natoms = fragment.number_of_atoms
+            for _ in range(count):
+                total.append(per_fragment + natoms)
+                natoms += f_natoms
+        return N.concatenate(total)
+
+    def atom_to_site_index_mapping(self):
+        """
+        :returns: an array whose element [s] is the site index
+                  corresponding to the first site of the atom with
+                  atom index s.
+        :rtype: numpy.ndarray
+        """
+        nsites = 0
+        total = []
+        for fragment, count in self.molecules:
+            per_fragment = fragment._atom_to_site_index_mapping()
+            for _ in range(count):
+                total.append(per_fragment + nsites)
+                nsites += per_fragment[-1]
+        m = N.concatenate(total)
+        m[1:] = m[:-1]
+        m[0] = 0
+        return m
+
+    def template_site_to_template_atom_index_mapping(self):
+        """
+        :returns: an array whose element [s] is the template atom index
+                  corresponding to template site index s.
+        :rtype: numpy.ndarray
+        """
+        natoms = 0
+        total = []
+        for fragment, count in self.molecules:
+            per_fragment = fragment.site_to_atom_index_mapping()
+            f_natoms = fragment.number_of_atoms
+            total.append(per_fragment + natoms)
+            natoms += f_natoms
+        return N.concatenate(total)
+
+    def site_to_template_index_mapping(self):
+        """
+        :returns: an array whose element [s] is the template site index
+                  corresponding to site index s.
+        :rtype: numpy.ndarray
+        """
+        ntsites = 0
+        total = []
+        for fragment, count in self.molecules:
+            f_nsites = fragment.number_of_sites
+            per_fragment = N.arange(f_nsites)
+            for _ in range(count):
+                total.append(per_fragment + ntsites)
+            ntsites += f_nsites
+        return N.concatenate(total)
+
+    def atom_to_template_index_mapping(self):
+        """
+        :returns: an array whose element [s] is the template atom index
+                  corresponding to atom index s.
+        :rtype: numpy.ndarray
+        """
+        ntatoms = 0
+        total = []
+        for fragment, count in self.molecules:
+            f_natoms = fragment.number_of_atoms
+            per_fragment = N.arange(f_natoms)
+            for _ in range(count):
+                total.append(per_fragment + ntatoms)
+            ntatoms += f_natoms
+        return N.concatenate(total)
+
 
 # Properties associate a value with each atom or site in a
 # universe. The value can be an array of any shape and any element
@@ -1027,6 +1158,18 @@ class MosaicConfiguration(MosaicDataItem):
                 'parallelepiped': lambda p: tuple(N.array(v) for v in p),
                 }[self.universe.cell_shape](self.cell_parameters)
 
+    def cell_volume(self):
+        """
+        :returns: the volume the simulation cell, or None for
+                  infinite universes
+        :rtype: float
+        """
+        return {'infinite': lambda p: None,
+                'cube': lambda p: float(p*p*p),
+                'cuboid': lambda p: p[0]*p[1]*p[2],
+                'parallelepiped': lambda p: abs(LA.det(p)),
+                }[self.universe.cell_shape](self.cell_parameters)
+
     # Equivalence test
 
     def validate_equivalence(self, other):
@@ -1095,6 +1238,151 @@ class MosaicConfiguration(MosaicDataItem):
                              "Configuration.positions must have same dtypes")
 
 
+# Selections specify a subset of atoms or sites, either in the templates
+# or in the molecules of a universe.
+
+class MosaicSelection(MosaicDataItem):
+
+    """Selection
+
+    See the :ref:`data model documentation<mosaic-selection>` for selections.
+
+    """
+
+    # API properties
+
+    @abstractproperty
+    def type(self):
+        """A string identifying the type of the selection.
+        See the :ref:`data model documentation<mosaic-selection-type>`.
+        """
+        raise NotImplementedError()
+
+    @abstractproperty
+    def universe(self):
+        "The :py:class:`MosaicUniverse` for which the selection is defined."
+        raise NotImplementedError()
+
+    @abstractproperty
+    def indices(self):
+        """An array containing the indices of the contained atoms or sites.
+        See the :ref:`data model documentation<mosaic-selection-indices>`.
+        """
+        raise NotImplementedError()
+
+    # Properties that can be computed in terms of the API properties
+
+    @property
+    def number_of_atoms(self):
+        """The number of atoms in the selection.
+        """
+        indices = self.indices
+        if self.type == "atom":
+            return len(indices)
+        elif self.type == "template_atom":
+            natoms = 0
+            ntatoms = 0
+            for fragment, count in self.universe.molecules:
+                nta = fragment.number_of_atoms
+                natoms += count * N.sum((indices >= ntatoms)
+                                        & (indices < ntatoms+nta))
+                ntatoms += nta
+            return natoms
+        else:
+            raise TypeError("number of atoms undefined in site selection")
+
+    @property
+    def number_of_sites(self):
+        """The number of sites in the selection.
+        """
+        indices = self.indices
+        if self.type == "atom":
+            sites_per_atom = [a.number_of_sites
+                              for a in self.universe.recursive_atom_iterator()]
+            # If 'indices' is an immutable array, N.take crashes due to
+            # a numpy bug. Conversion to a plain array prevents this.
+            # See Github issue #3758 for numpy/numpy.
+            return N.sum(N.take(sites_per_atom, N.array(indices)))
+        elif self.type == "template_atom":
+            sites_per_atom = [a.number_of_sites
+                              for a in self.universe.recursive_atom_iterator()]
+            nsites = 0
+            ntatoms = 0
+            for fragment, count in self.universe.molecules:
+                nta = fragment.number_of_atoms
+                mask = (indices >= ntatoms) & (indices < ntatoms+nta)
+                ntatoms += nta
+                # Conversion to plain arrays is required for the same
+                # reason as above.
+                nsites += count * N.sum(N.take(sites_per_atom,
+                                               N.repeat(N.array(indices),
+                                                        N.array(mask))))
+            return nsites
+        elif self.type == "site":
+            return len(indices)
+        elif self.type == "template_site":
+            nsites = 0
+            ntsites = 0
+            for fragment, count in self.universe.molecules:
+                nts = fragment.number_of_sites
+                nsites += count * N.sum((indices >= ntsites)
+                                        & (indices < ntsites+nts))
+                ntsites += nts
+            return nsites
+
+    # Equivalence test
+
+    def validate_equivalence(self, other):
+        """Verify the equivalence of Python objects representing
+        Mosaic data. The two objects can belong to different models;
+        only the common API functionality is used for the test.
+
+        :parameter other: an arbitrary Python object
+        :raises ValueError: if other is not equivalent to self
+        """
+        if not isinstance(other, MosaicSelection):
+            raise TypeError("%s is not a selection" % str(type(other)))
+        self.universe.validate_equivalence(other.universe)
+        if self.type != other.type:
+            raise ValueError("types differ: %s != %s"
+                             % (repr(self.type), repr(other.type)))
+        if (self.indices != other.indices).any():
+            raise ValueError("indices differ")
+
+    def is_equivalent(self, other):
+        """Check for equivalence of Python objects representing
+        Mosaic data. The two objects can belong to different models;
+        only the common API functionality is used for the test.
+
+        :parameter other: an arbitrary Python object
+        :returns: True if other is equivalent to self
+        :rtype: bool
+        """
+        try:
+            self.validate_equivalence(other)
+            return True
+        except:
+            return False
+
+    # Validation
+
+    _allowed_types = ["atom", "site", "template_atom", "template_site"]
+
+    def validate(self):
+        """Verify that the object satisfies the constraints of the
+        Mosaic data model.
+
+        :raises ValueError: if the object is not valid Mosaic data
+        """
+        validate_value(self.type, self._allowed_types, "Selection.type")
+        validate_type(self.universe, MosaicUniverse, "Selection.universe")
+        max_index = {"atom": self.universe.number_of_atoms,
+                     "site": self.universe.number_of_sites,
+                     "template_atom": self.universe.number_of_template_atoms,
+                     "template_site": self.universe.number_of_template_sites,
+                    }[self.type]
+        validate_indices(self.indices, max_index, "Selection.indices")
+
 # Validation functions
 
 def validate_type(obj, cls, text):
@@ -1156,6 +1444,18 @@ def validate_sequence(obj, el_cls, text, additional_tests=()):
             if not test_fn(el):
                 raise ValueError("%s: %s" % (str(el), text))
 
+def validate_indices(indices, max_index, text):
+    validate_array(indices, None,
+                   [N.uint8, N.uint16, N.uint32, N.uint64],
+                   text)
+    if len(indices.shape) != 1:
+        raise ValueError("index array not 1d")
+    if (indices >= max_index).any():
+        raise ValueError("index too large")
+    if len(indices) > 1:
+        d = indices[1:]-indices[:-1]
+        if (d <= 0).any():
+            raise ValueError("indices not sorted")
 
 # The unit validator accepts a very limited syntax, which
 # should be sufficient: a unit is defined by a string of
@@ -1167,16 +1467,31 @@ def validate_sequence(obj, el_cls, text, additional_tests=()):
 
 def validate_units(unit_spec, text):
     validate_type(unit_spec, str, text)
+    first = True
     for unit in unit_spec.split():
-        if not (re.match("^[a-zA-Z]+([-]?[0-9]+)?$", unit)
-                # unit symbol plus exponent
-                or re.match("^[1-9][0-9]*$", unit)
-                # integer
-                or re.match("^0\.[0-9]+$", unit)
-                # decimal fraction
-                ):
-            raise ValueError("invalid unit '%s' in %s" % (unit, unit_spec))
+        # number (integer or decimal fraction)
+        if first:
+            first = False
+            if (re.match("^[1-9][0-9]*$", unit) \
+                or re.match("^0\.[0-9]+$", unit)):
+                continue
+        # symbol+exponent
+        m = re.match("^(?P<symbol>[a-zA-Z]+)([-]?[0-9]+)?$", unit)
+        if m and m.group('symbol') in _allowed_units:
+            continue
+        raise ValueError("invalid unit '%s' in %s" % (unit, unit_spec))
 
+_allowed_units = \
+    ["pm",   "Ang",  "nm",   "um",   "mm",   "m",    
+     "fs",   "ps",   "ns",   "us",   "ms",   "s",    
+     "amu",  "g",    "kg",   
+     "mol",  
+     "J",    "kJ",   "cal",  "kcal", "eV",   
+     "K",    
+     "Pa",   "kPa",  "MPa",  "GPa",  "atm",  "bar",  "kbar", 
+     "e",    "C",    "A",    "V",    
+     "rad",  
+     "c",    "h",    "me"]
 
 # Validation as a test function rather than raising exceptions
 
