@@ -1,16 +1,42 @@
 # -*- coding: utf-8 -*-
 """HDF5 I/O
 
+.. moduleauthor:: Konrad Hinsen
+
 Any model implementing the Mosaic API can be written to a HDF5 file,
-but will be implicitly converted to an array_model representation.
-Data items read from a HDF5 file are always represented by
-an array_model.
+but will be implicitly converted to a :mod:`mosaic.array_model`
+representation.  Data items read from a HDF5 file are always
+represented by an :mod:`mosaic.array_model`.
 
 The HDF5 storage layout is designed for clarity rather than
 simplicity, the files should be as self-documenting as possible.
 
+A common pattern for generating an HDF5 file is:
+
+::
+
+    with HDF5Store('molecule.h5', 'w') as writer:
+       writer.store("universe", universe)
+       writer.store("configuration1", configuration1)
+       writer.store("configuration2", configuration2)
+
+A common pattern for reading from an HDF5 file is:
+
+::
+
+    items = {}
+    for id, data in HDF5Store('molecule.h5'):
+       items[id] = data
+
 """
 
+#-----------------------------------------------------------------------------
+#       Copyright (C) 2013 The Mosaic Development Team
+#
+#  Distributed under the terms of the BSD License.  The full license is in
+#  the file LICENSE.txt, distributed as part of this software.
+#-----------------------------------------------------------------------------
+
 import copy
 import operator
 import weakref
@@ -37,27 +63,41 @@ if h5py.version.hdf5_version_tuple < (1, 8, 7):
     raise ValueError("HDF5 version >= 1.8.7 required")
 
 
-def int_size_for_max_value(n):
-    if n < 256:
-        return N.uint8
-    elif n < 65536:
-        return N.uint16
-    elif n < 4294967296:
-        return N.uint32
-    else:
-        return N.uint64
-
-
 def create_dataset(group, path, data):
-    ds = group.require_dataset(path, shape=data.shape, dtype=data.dtype)
-    if 0 not in data.shape:
+    if isstring(data):
+        ds = group.require_dataset(path, shape=(),
+                                   dtype=h5py.special_dtype(vlen=bytes))
         ds[...] = data
+    else:
+        ds = group.require_dataset(path, shape=data.shape, dtype=data.dtype)
+        if 0 not in data.shape:
+            ds[...] = data
     return ds
 
 
 class HDF5Store(object):
 
+    """I/O handler for HDF5 files
+
+    This class handles the translation of references between Mosaic
+    data items in memory and HDF5 object references in a file. When
+    writing data, references are allowed only to data items that have
+    been written earlier using the same HDF5Store instance.
+
+    An HDF5Store can be used like a file, in which case it must be
+    closed after the last ``store`` or ``retrieve`` operation, or as a
+    context manager in a with-statement. For reading, it also behaves
+    like an iterator yielding ``(name, data_item)`` pairs.
+    """
+
     def __init__(self, hdf5_group_or_filename, file_mode=None):
+        """
+        :param hdf5_group_or_filename: a group in an open HDF5 file,
+                                       or a string interpreted as a file name
+        :type hdf5_group_or_filename: str or ``h5py.Group``
+        :param file_mode: ``'r'`` or ``'w'``, used only with a filename argument
+        :type file_mode:  str
+        """
         if isstring(hdf5_group_or_filename):
             if file_mode is None:
                 file_mode = 'r'
@@ -89,6 +129,12 @@ class HDF5Store(object):
     storage_handler = MethodRegister()
 
     def store(self, path, data):
+        """
+        :param path: a HDF5 path, relative to the group used by the HDF5Store
+        :type path:  str
+        :param data: a Mosaic data item
+        :type data:  :class:`mosaic.api.MosaicDataItem`
+        """
         api.validate_type(path, str, "path")
         if path[0] == "/":
             raise ValueError("absolute paths not allowed")
@@ -101,6 +147,13 @@ class HDF5Store(object):
         self._register_data_item(path, data)
 
     def retrieve(self, path_or_node):
+        """
+        :param path_or_node: a HDF5 path, relative to the group used by
+                             the HDF5Store, or an HDF5 node
+        :type path_or_node:  str or h5py.Node
+        :returns: a Mosaic data item
+        :rtype:  :class:`mosaic.api.MosaicDataItem`
+        """
         if isinstance(path_or_node, h5py.h5r.Reference):
             path_or_node = self.root[path_or_node]
         if isinstance(path_or_node, str):
@@ -129,22 +182,45 @@ class HDF5Store(object):
         self._register_data_item(path, data)
         return data
 
-    _node_types = ['universe', 'configuration', 'property', 'label']
+    _node_types = ['universe', 'configuration', 'property', 'label',
+                   'selection']
 
     def __iter__(self):
         # First pass: universes
         for name in self.root:
             node = self.root[name]
-            if node.attrs['DATA_MODEL'] == ascii('MOSAIC') \
+            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
                and node.attrs['MOSAIC_DATA_TYPE'] == ascii('universe'):
                 yield name, self.retrieve(node)
-        # First pass: anything else
+        # Second pass: anything else
         for name in self.root:
             node = self.root[name]
-            if node.attrs['DATA_MODEL'] == ascii('MOSAIC') \
+            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
                and node.attrs['MOSAIC_DATA_TYPE'] != ascii('universe'):
                 yield name, self.retrieve(node)
 
+    def recursive_iterator(self, group=None):
+        if group is None:
+            group = self.root
+        # First pass: universes
+        for name in group:
+            node = group[name]
+            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
+               and node.attrs['MOSAIC_DATA_TYPE'] == ascii('universe'):
+                yield name, self.retrieve(node)
+            if isinstance(node, h5py.Group):
+                for gname, gnode in self.recursive_iterator(node):
+                    yield name+'/'+gname, gnode
+        # Second pass: anything else
+        for name in group:
+            node = group[name]
+            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
+               and node.attrs['MOSAIC_DATA_TYPE'] != ascii('universe'):
+                yield name, self.retrieve(node)
+            if isinstance(node, h5py.Group):
+                for gname, gnode in self.recursive_iterator(node):
+                    yield name+'/'+gname, gnode
+
     def _stamp(self, node, label):
         node.attrs['DATA_MODEL'] = ascii('MOSAIC')
         node.attrs['DATA_MODEL_MAJOR_VERSION'] = api.MOSAIC_VERSION[0]
@@ -191,11 +267,11 @@ class HDF5Store(object):
 
         # Create HDF5 datasets
         create_dataset(group, 'cell_shape',
-                       data=N.array(universe.cell_shape, 'S'))
+                       data=universe.cell_shape)
         create_dataset(group, 'symmetry_transformations',
                        data=N.array(universe._symmetry_transformations))
         create_dataset(group, 'convention',
-                       data=N.array(universe.convention, 'S'))
+                       data=universe.convention)
         create_dataset(group, 'fragments', data=universe._fragment_array)
         create_dataset(group, 'atoms', data=universe._atom_array)
         create_dataset(group, 'bonds', data=universe._bond_array)
@@ -278,6 +354,18 @@ class HDF5Store(object):
         ds.attrs['name'] = ascii(label.name)
         ds.attrs['label_type'] = ascii(label.type)
 
+    @storage_handler(api.MosaicSelection)
+    def _store_selection(self, path, selection):
+        universe = selection.universe
+        universe_path = self._get_path(universe)
+        if universe_path is None:
+            raise IOError("universe must be stored first")
+        ds = create_dataset(self.root, path,
+                            N.ascontiguousarray(selection.indices))
+        ds.attrs['universe'] = self.root[universe_path].ref
+        ds.attrs['selection_type'] = ascii(selection.type)
+        self._stamp(ds, 'selection')
+
     #
     # Retrieve data from HDF5 file
     #
@@ -298,8 +386,8 @@ class HDF5Store(object):
                         node['molecules'][...],
                         tuple(py_str(s) for s in node['symbols'][...]),
                         st,
-                        py_str(node['cell_shape'][...]),
-                        py_str(node['convention'][...]))
+                        py_str(node['cell_shape'][()]),
+                        py_str(node['convention'][()]))
 
     @storage_handler('configuration')
     def _retrieve_configuration(self, node):
@@ -334,3 +422,11 @@ class HDF5Store(object):
         label_type = py_str(node.attrs['label_type'])
         return mosaic.array_model.Label.from_arrays(universe, name,
                                                     data, label_type)
+
+    @storage_handler('selection')
+    def _retrieve_selection(self, node):
+        universe = self.retrieve(node.attrs['universe'])
+        indices = node[...]
+        selection_type = py_str(node.attrs['selection_type'])
+        return mosaic.array_model.Selection(universe, indices,
+                                            selection_type)
