@@ -1,8 +1,41 @@
 # -*- coding: utf-8 -*-
 """XML I/O
 
+.. moduleauthor:: Konrad Hinsen
+
+The XML representation of Mosaic data consists of a single element
+``<mosaic version="x.y">...</mosaic>`` that can contain any number of
+Mosaic data items, each of which is identified by a unique id.
+The classes :class:`XMLWriter` and :class:`XMLReader` handle the
+translation between Mosaic data items in memory that can reference
+each other and XML elements that reference each other by id.
+
+A common pattern for generating an XML file is:
+
+::
+
+    with XMLWriter('molecule.xml') as writer:
+       writer.store("universe", universe)
+       writer.store("configuration1", configuration1)
+       writer.store("configuration2", configuration2)
+
+A common pattern for reading from an XML file is:
+
+::
+
+    items = {}
+    for id, data in XMLReader('molecule.xml'):
+       items[id] = data
+
 """
 
+#-----------------------------------------------------------------------------
+#       Copyright (C) 2013 The Mosaic Development Team
+#
+#  Distributed under the terms of the BSD License.  The full license is in
+#  the file LICENSE.txt, distributed as part of this software.
+#-----------------------------------------------------------------------------
+
 try:
     import xml.etree.cElementTree as ET
 except IOError:
@@ -14,6 +47,7 @@ import immutable.np as IN
 import mosaic.api as api
 import mosaic.immutable_model as im
 from mosaic.utility import MethodRegister
+from mosaic.utility import uint_for_max_value
 from mosaic.utility import isstring
 from mosaic.utility import xml_encoding
 
@@ -61,7 +95,6 @@ def s2t(s):
             "boolean": N.bool
             }[s]
 
-
 class XMLStore(object):
 
     def __init__(self):
@@ -89,7 +122,24 @@ class XMLStore(object):
 
 class XMLWriter(XMLStore):
 
+    """Output handler for XML files
+
+    This class handles the translation of references between Mosaic
+    data items in memory to XML references by id. References are
+    allowed only to data items that have been written earlier using
+    the same XMLWriter instance.
+
+    An XMLWriter can be used like a file, in which case it must be
+    closed after the last ``store`` operation, or as a context manager
+    in a with-statement.
+    """
+
     def __init__(self, xml_file):
+        """
+        :param xml_file: a writeable file object, or a string
+                         interpreted as a file name
+        :type xml_file: str or file-like
+        """
         XMLStore.__init__(self)
         if isstring(xml_file):
             self.file = open(xml_file, 'w')
@@ -120,6 +170,12 @@ class XMLWriter(XMLStore):
     storage_handler = MethodRegister()
 
     def store(self, xml_id, data):
+        """
+        :param xml_id: the id of the XML element representing the data item
+        :type xml_id:  str
+        :param data:   a Mosaic data item
+        :type data:    :class:`mosaic.api.MosaicDataItem`
+        """
         api.validate_type(xml_id, str, "xml_id")
         api.validate_type(data, api.MosaicDataItem, "data")
         xml_id = xml_id
@@ -214,7 +270,7 @@ class XMLWriter(XMLStore):
                                   for p in configuration.cell_parameters.flat)
             el.append(el_cp)
         el_pos = ET.Element('positions', type=float_type)
-        el_pos.text = '|'.join(' '.join(n2s(x, dp) for x in p)
+        el_pos.text = ' '.join(' '.join(n2s(x, dp) for x in p)
                                for p in configuration.positions)
         el.append(el_pos)
         self._write(el)
@@ -235,7 +291,7 @@ class XMLWriter(XMLStore):
                              shape=shape_str,
                              type=t2s(property.data.dtype))
         dp = property.data.dtype == N.dtype(N.float64)
-        el_data.text = '|'.join(' '.join(n2s(x, dp) for x in v.flat)
+        el_data.text = ' '.join(' '.join(n2s(x, dp) for x in v.flat)
                                 for v in property.data)
         el.append(el_data)
         self._write(el)
@@ -255,14 +311,41 @@ class XMLWriter(XMLStore):
         el.append(el_strings)
         self._write(el)
 
+    @storage_handler(api.MosaicSelection)
+    def _store_selection(self, xml_id, selection):
+        universe = selection.universe
+        universe_id = self._get_id(universe)
+        if universe_id is None:
+            raise IOError("universe must be stored first")
+        el = ET.Element(selection.type + '_selection')
+        el.set("id", xml_id)
+        el.append(ET.Element('universe', ref=universe_id))
+        el_indices = ET.Element("indices")
+        el_indices.text = ' '.join(str(i) for i in selection.indices)
+        el.append(el_indices)
+        self._write(el)
+
 
 class XMLReader(XMLStore):
 
-    # The XML reader does not do any validation, it assumes the XML
-    # input to be correct. Validation will be left for later, when we
-    # have a schema.
+    """Input handler for XML files
+
+    This class handles the translation of references by id in an XML
+    file to in-memory references between data items.
+
+    An XMLReader is used as an iterator over ``(id, data_item)`` pairs.
+
+    The current implementation does not do any validation, it assumes
+    the XML input to be correct.
+    """
+
 
     def __init__(self, xml_file):
+        """
+        :param xml_file: a file object, or a string
+                         interpreted as a file name
+        :type xml_file: str or file-like
+        """
         XMLStore.__init__(self)
         if isstring(xml_file):
             # file name given: open file and close it at the end
@@ -278,10 +361,14 @@ class XMLReader(XMLStore):
             self.file.close()
         # No more reading after this
         self.file = None
+        self._close_file = False
         # Clear cache
         self._id_map = None
         self._data_map = None
 
+    def __del__(self):
+        self.close()
+
     def __iter__(self):
         for event, el in ET.iterparse(self.file):
             if el.tag == 'mosaic':
@@ -400,6 +487,22 @@ class XMLReader(XMLStore):
         strings = tuple(el_strings.text.split())
         return klass(universe, el.get('name'), strings)
 
+    @data_handler("atom_selection")
+    @data_handler("site_selection")
+    @data_handler("template_atom_selection")
+    @data_handler("template_site_selection")
+    def _read_selection(self, el):
+        universe = self._read_universe(el.find('universe'))
+        ptype = el.tag[:-10]   # strip off '_selection'
+        klass = {"atom": im.AtomSelection,
+                 "site": im.SiteSelection,
+                 "template_atom": im.TemplateAtomSelection,
+                 "template_site": im.TemplateSiteSelection}[ptype]
+        el_d = el.find('indices')
+        indices = [int(s) for s in el_d.text.split()]
+        indices = IN.array(indices, uint_for_max_value(max(indices)))
+        return klass(universe, indices)
+
     # Parse array data
 
     def _array(self, shape, dtype, text):
@@ -408,8 +511,10 @@ class XMLReader(XMLStore):
         return IN.array([dtype(s) for s in text.split()]).reshape(shape)
 
     def _property(self, el_shape, dtype, text):
+        data = IN.array([dtype(s) for s in text.split()])
         if isstring(el_shape):
             el_shape = tuple(int(s) for s in el_shape.split())
-        return IN.array(
-            tuple(IN.array([dtype(s) for s in part.split()]).reshape(el_shape)
-                  for part in text.split('|')))
+        n_els = N.product(el_shape, dtype=N.int)
+        assert len(data) % n_els == 0
+        n_entries = len(data) // n_els
+        return data.reshape((n_entries,)+el_shape)
