@@ -26,6 +26,7 @@ from collections import OrderedDict
 import copy
 import itertools as it
 import os
+import re
 import sys
 import traceback
 
@@ -37,7 +38,7 @@ from mosaic_pdb.space_groups import space_groups
 
 from mosaic_pdb.pdb_chem_comp import components, variants
 from mosaic.mutable_model import \
-        Element, Atom, Fragment, Polymer, Universe, \
+        Element, Unknown, Atom, Fragment, Polymer, Universe, \
         Configuration, SiteProperty
 from mosaic import api
 
@@ -49,6 +50,17 @@ class Units:
     Ang = 0.1
     deg = np.pi / 180.
 
+# Bond order translation
+
+def bond_order(pdb_value_order):
+    if pdb_value_order is None:
+        return ''
+    else:
+        return {'sing': 'single',
+                'doub': 'double',
+                'trip': 'triple',
+                'quad': 'qudruple',
+                'arom': 'aromatic'}[pdb_value_order.lower()]
 
 # Crystallographic unit cell
 
@@ -77,7 +89,8 @@ class UnitCell(object):
 
 class MMCIFStructure(object):
 
-    def __init__(self, structure_file=None, pdb_code=None):
+    def __init__(self, structure_file=None, pdb_code=None,
+                 with_auth_spec=False):
         """
         Specify the data to be loaded. The following combinations
         are valid:
@@ -92,6 +105,10 @@ class MMCIFStructure(object):
         @type structure_file: C{str}
         @param pdb_code: a four-letter PDB code
         @type pdb_code: C{str}
+        @param with_auth_spec: flag for reading and storing the
+                               author's alternative atom site
+                               specifications
+        @type with_auth_spec: C{bool}
         """
         if pdb_code is not None:
             self.pdb_code = pdb_code
@@ -107,7 +124,6 @@ class MMCIFStructure(object):
         else:
             raise MMCIFError("No input structure given")
 
-        self.experiment = {}
         self.cell_parameters = {}
         self.symmetry = {}
         self.entities = OrderedDict()
@@ -117,10 +133,14 @@ class MMCIFStructure(object):
         self.extra_bonds = []
         self.models = []
         self.u = {}
+        if with_auth_spec:
+            self.auth_spec = {}
+        else:
+            self.auth_spec = None
         self.modified = False
         parser.parseToObjects(cell=self.cell_parameters,
                               symmetry=self.symmetry,
-                              exptl=self.experiment,
+                              exptl=self.getPDBCodeFromExptl,
                               entity=self.addEntity,
                               entity_name_com=self.addEntityName,
                               entity_poly=self.addPolymerInfo,
@@ -133,18 +153,19 @@ class MMCIFStructure(object):
                               atom_site_anisotrop=self.addAnisoU)
         if self.structure_file is not None:
             self.structure_file.close()
-        if 'entry_id' in self.experiment:
-            if self.pdb_code is None:
-                self.pdb_code = self.experiment['entry_id']
-            else:
-                assert self.pdb_code == self.experiment['entry_id']
         if len(self.cell_parameters) == 0 \
                or (self.cell_parameters['length_a'] ==
                    self.cell_parameters['length_b'] ==
                    self.cell_parameters['length_a'] == '1'
                    and self.cell_parameters['angle_alpha'] ==
                        self.cell_parameters['angle_beta'] ==
-                       self.cell_parameters['angle_gamma'] == '90'):
+                       self.cell_parameters['angle_gamma'] == '90') \
+               or (self.cell_parameters['length_a'] ==
+                   self.cell_parameters['length_b'] ==
+                   self.cell_parameters['length_a'] == '?'
+                   and self.cell_parameters['angle_alpha'] ==
+                       self.cell_parameters['angle_beta'] ==
+                       self.cell_parameters['angle_gamma'] == '?'):
             # NMR or other non-crystal structure
             self.cell = None
         else:
@@ -187,6 +208,13 @@ class MMCIFStructure(object):
         assert value in [None, '.']
         return None
 
+    def getPDBCodeFromExptl(self, indices, data):
+        entry_id = self.getField('entry_id', indices, data)
+        if self.pdb_code is None:
+            self.pdb_code = entry_id
+        else:
+            assert self.pdb_code == entry_id
+
     def addEntity(self, indices, data):
         data = dict((label, data[index])
                     for label, index in indices.items()
@@ -213,10 +241,19 @@ class MMCIFStructure(object):
         self.entities[entity_id]['sequence'] = seq
         res_number = int(data[indices['num']])
         res_type = data[indices['mon_id']]
-        if seq and res_number - seq[-1][0] != 1:
-            raise ValueError("non-consecutive residue numbers "
-                             "in entity_poly_seq")
-        seq.append((res_number, res_type))
+        if self.getYesNoField('hetero', indices, data):
+            if seq and res_number == seq[-1][0]:
+                # Consecutive monomer at the same position as the
+                # preceding one.
+                seq[-1] = (res_number, seq[-1][1] + (res_type,))
+            else:
+                # First monomer at this position.
+                seq.append((res_number, (res_type,)))
+        else:
+            if seq and res_number - seq[-1][0] != 1:
+                raise ValueError("non-consecutive residue numbers %d-%d"
+                                 "in entity_poly_seq" % (res_number, seq[-1][0]))
+            seq.append((res_number, res_type))
 
     def addMolecule(self, indices, data):
         data = dict((label, data[index])
@@ -232,22 +269,47 @@ class MMCIFStructure(object):
 
     def addConnection(self, indices, data):
         if self.getField('conn_type_id', indices, data) == 'disulf':
-            asym1, comp1, seq1, asym2, comp2, seq2 = \
+            asym1, comp1, seq1, sym1, asym2, comp2, seq2, sym2 = \
                    [self.getField(f, indices, data)
                     for f in ['ptnr1_label_asym_id', 'ptnr1_label_comp_id',
-                              'ptnr1_label_seq_id', 'ptnr2_label_asym_id',
-                              'ptnr2_label_comp_id', 'ptnr2_label_seq_id']]
-            self.ss_bridges.append(((asym1, comp1, int(seq1)),
-                                    (asym2, comp2, int(seq2))))
-        elif self.getField('conn_type_id', indices, data) == 'covale':
-            asym1, comp1, seq1, atom1, asym2, comp2, seq2, atom2 = \
-                   [self.getField(f, indices, data)
-                    for f in ['ptnr1_label_asym_id', 'ptnr1_label_comp_id',
-                              'ptnr1_label_seq_id', 'ptnr1_label_atom_id',
+                              'ptnr1_label_seq_id',  'ptnr1_symmetry',
                               'ptnr2_label_asym_id', 'ptnr2_label_comp_id',
-                              'ptnr2_label_seq_id', 'ptnr2_label_atom_id']]
+                              'ptnr2_label_seq_id',  'ptnr2_symmetry']]
+            if sym1.split('_')[0] == 1 and sym2.split('_')[0] == 1:
+                self.ss_bridges.append(((asym1, comp1, int(seq1)),
+                                        (asym2, comp2, int(seq2))))
+            else:
+                # For now, we ignore disulfide bonds between symmetry-related
+                # chains. The only way to handle this in MOSAIC would be
+                # to generate symmetry equivalents explicitly.
+                pass
+        elif self.getField('conn_type_id', indices, data) == 'covale':
+            bond = ()
+            for f in ['ptnr1_label_asym_id', 'ptnr1_label_comp_id',
+                      ('ptnr1_label_seq_id', 'ptnr1_auth_seq_id'),
+                      'ptnr1_label_atom_id',
+                      'ptnr2_label_asym_id', 'ptnr2_label_comp_id',
+                      ('ptnr2_label_seq_id', 'ptnr2_auth_seq_id'),
+                      'ptnr2_label_atom_id',
+                      'pdbx_value_order']:
+                if isinstance(f, tuple):
+                    value = None
+                    for ff in f:
+                        value = self.getField(ff, indices, data)
+                        if value not in [None, '.']:
+                            break
+                    if value in [None, '.']:
+                        raise ValueError('Field %s has invalid value %s'
+                                         % (ff, value))
+                    else:
+                        bond += (value,)
+                else:
+                    bond += (self.getField(f, indices, data),)
+            asym1, comp1, seq1, atom1, asym2, comp2, seq2, atom2, \
+                value_order  = bond
             self.extra_bonds.append(((asym1, comp1, seq1, atom1),
-                                     (asym2, comp2, seq2, atom2)))
+                                     (asym2, comp2, seq2, atom2),
+                                     value_order))
 
     def recordAltLabel(self, indices, data):
         data = dict((label, data[index])
@@ -269,8 +331,9 @@ class MMCIFStructure(object):
         entity = self.entities[self.molecules[asym_id]['entity_id']]
         is_polymer = entity['type']  == 'polymer' \
                      and 'polymer_type' in entity
+        is_multimer = entity['type']  == 'polymer' \
+                      and 'polymer_type' not in entity
         if is_polymer:
-            seq_id_string = data[indices['label_seq_id']] 
             try:
                 seq_id = int(data[indices['label_seq_id']])
             except ValueError:
@@ -280,6 +343,24 @@ class MMCIFStructure(object):
                 raise ValueError('invalid seq_id %d for atom id %s'
                                  % (seq_id, unique_id))
             ins_code = self.getField('pdbx_PDB_ins_code', indices, data)
+        elif is_multimer:
+            # These are mostly sugars. The consist of several
+            # monomers, which however have no seq_id because in
+            # general they are not linear. The only way to uniquely
+            # identify the monomers at this time is the auth_seq_id
+            # field. The PDB is working on a better scheme.
+            try:
+                seq_id = int(data[indices['label_seq_id']])
+            except ValueError:
+                try:
+                    seq_id = int(data[indices['auth_seq_id']])
+                except ValueError:
+                    raise ValueError('non-numeric label_seq_id %s '
+                                     'and auth_seq_id %s for atom id %s'
+                                     % (data[indices['label_seq_id']],
+                                        data[indices['auth_seq_id']],
+                                        unique_id))
+            ins_code = self.getField('pdbx_PDB_ins_code', indices, data)
         else:
             seq_id = None
             ins_code = None
@@ -291,7 +372,7 @@ class MMCIFStructure(object):
         while len(self.models) < model_number:
             self.models.append(OrderedDict())
         model = self.models[model_number - 1]
-        if is_polymer:
+        if is_polymer or is_multimer:
             molecule = model.get(asym_id, [[]])
             monomer = molecule[-1]
             if len(monomer) > 0 and \
@@ -316,6 +397,16 @@ class MMCIFStructure(object):
         monomer.append((unique_id, atom_id, alt_id, comp_id, seq_id, ins_code,
                         atom_type, x, y, z, occupancy, u_iso))
 
+        if self.auth_spec is not None:
+            auth_asym_id = self.getField('auth_asym_id', indices, data)
+            auth_atom_id = self.getField('auth_atom_id', indices, data)
+            auth_comp_id = self.getField('auth_comp_id', indices, data)
+            auth_seq_id = self.getField('auth_seq_id', indices, data)
+            self.auth_spec[unique_id] =  dict(asym_id = auth_asym_id,
+                                              atom_id = auth_atom_id,
+                                              comp_id = auth_comp_id,
+                                              seq_id  = int(auth_seq_id))
+
         stop = False
         for field in ['adp_type', 'aniso_B[1][1]', 'aniso_U[1][1]']:
             value = self.getField(field, indices, data)
@@ -364,81 +455,129 @@ def partition_sites(sites):
 def make_monomer(sites, site_properties):
     comp_ids = [s[3] for s in sites]
     if len(set(comp_ids)) > 1:
-        raise ValueError("Can't handle monomer made of multiple components: %s"
-                         % str(set(comp_ids)))
-        if False:
-            fragments = []
-            unique_ids = []
-            comp_ids = []
-            label_parts = []
-            for i, (comp_id, group) in enumerate(partition_sites(sites)):
-                f, u = make_monomer(group, site_properties)
-                comp_ids.append(comp_id)
-                label_parts.append(comp_id + '_' + str(i+1))
-                fragments.append(f)
-                unique_ids.extend(u)
-            bonds = ()  # !!! TODO
-            return Fragment('+'.join(label_parts), '+'.join(comp_ids),
-                            fragments, (), bonds), \
-                   unique_ids       
+        # Make a heterogeneous monomer
+        ms = [_make_monomer(comp_id, list(comp_sites), site_properties)
+              for comp_id, comp_sites in it.groupby(sites, lambda s: s[3])]
+        fs = [f for f, u in ms]
+        us = [u for f, u in ms]
+        species = '/'.join(f.label.split('_')[0] for f in fs)
+        seq_id = fs[0].label.split('_')[1]
+        return Fragment(species+'_'+seq_id, species, fs, (), ()), sum(us, ())
+    else:
+        # Make a standard monomer
+        return _make_monomer(comp_ids[0], sites, site_properties)
+
+def make_multimer(asym_id, sites, site_properties, extra_bonds):
+    # Polymers without a polymer type. Make individual monomers and
+    # assemble them into a fragment. Check for covalent bonds between
+    # the monomers.
+    ms = [_make_monomer(monomer_sites[0][3], monomer_sites, site_properties)
+          for monomer_sites in sites]
+    fs = [f for f, u in ms]
+    us = [u for f, u in ms]
+    bonds = [("%s_%s.%s" % (a1[1], a1[2], a1[3]),
+              "%s_%s.%s" % (a2[1], a2[2], a2[3]),
+              bond_order(value_order))
+             for a1, a2, value_order in extra_bonds
+             if a1[0] == asym_id and a2[0] == asym_id]
+    species = '-'.join(f.label.split('_')[0] for f in fs)
+    return Fragment(species, species, fs, (), bonds), sum(us, ())
+
+def _make_monomer(comp_id, sites, site_properties):
+    # Entries from neutron scattering contain atoms of type D,
+    # whose names have a D in place of the H in the corresponding
+    # hydrogens in comp_id. These Ds must be translated to Hs in order
+    # to find the bonds. In the MOSAIC universe, the names starting
+    # with D are kept, but the element is set to H because D is not
+    # a valid chemical element in MOSAIC.
+    mosaic_ids = {}
+    for i in range(len(sites)):
+        s = sites[i]
+        atom_id = s[1]
+        mosaic_ids[atom_id] = atom_id
+        if s[6] == 'D':
+            if atom_id.startswith('D'):
+                atom_id = 'H' + atom_id[1:]
+            mosaic_ids[atom_id] = s[1]
+            sites[i] = (s[0], atom_id) + s[2:6] + ('H',) + s[7:]
+    # Heavy water has its own chemical component, DOD, with atoms
+    # of type D. In order to make the special treatment above work
+    # for this, DOD is replaced by HOH.
+    if comp_id == 'DOD':
+        comp_id = 'HOH'
+
     atom_ids = set(s[1] for s in sites)
-    comp = components[comp_ids[0]]
+    comp = components[comp_id]
     comp_atom_ids = set(comp.atoms['atom_id'])
+    species = comp.id
     if hasattr(comp, 'bonds'):
-        bond_orders = {'sing': 'single',
-                       'doub': 'double',
-                       'trip': 'triple',
-                       'quad': 'qudruple',
-                       'arom': 'aromatic'}
-        bonds = tuple((x['atom_id_1'], x['atom_id_2'],
-                       bond_orders.get(x['value_order'].lower(), ''))
+        bonds = tuple((mosaic_ids[x['atom_id_1']], mosaic_ids[x['atom_id_2']],
+                       bond_order(x.get('value_order')))
                       for x in comp.bonds
                       if x['atom_id_1'] in atom_ids
                          and x['atom_id_2'] in atom_ids)
     else:
         bonds = ()
-    if not comp_atom_ids.issuperset(atom_ids):
-        if atom_ids.difference(comp_atom_ids) == set(["HO5'"]) \
-           and "O5'" in atom_ids:
-            # Provide a quick fix for a common case in the PDB, to avoid
-            # having lots of DNA structures fail for a trivial reason
-            comp_atom_ids.add("HO5'")
-            bonds = bonds + (("HO5'", "O5'", 'single'),)
-        else:
-            # Variants exist only for amino acids. Find those that match
-            # the atom_ids in the current monomer.
-            vs = [v for v in variants.get(comp_ids[0], [])
-                  if set(components[v].atoms['atom_id']).issuperset(atom_ids)]
-            if not vs:
-                missing = atom_ids.difference(comp_atom_ids)
-                raise ValueError("Atom(s) %s not in component %s"
-                                 % (', '.join(missing), comp_ids[0]))
-            variant = None
-            if len(vs) == 1:
-                variant = vs[0]
-            else:
-                diff = [len(set(components[v].atoms['atom_id'])
-                            .difference(atom_ids))
-                        for v in vs]
-                if sum(d == 0 for d in diff) == 1:
-                    variant = vs[diff.index(0)]
-                else:
-                    seq_id = sites[0][4]
-                    if seq_id == 1:
-                        # N-terminal variants have LSN3 in their name
-                        vs = [v for v in vs if 'LSN3' in v]
-                        if len(vs) == 1:
-                            variant = vs[0]
-                        else:
-                            diff = [len(set(components[v].atoms['atom_id'])
-                                        .difference(atom_ids))
-                                    for v in vs]
-                            if sum(d == 0 for d in diff) == 1:
-                                variant = vs[diff.index(0)]
-            if variant is None:
-                raise ValueError("Monomer variant not unique: %s" % str(vs))
-            comp = components[variant]
+
+    # This part deals with monomers that contain atoms which are
+    # not in the component dictionary entry. If these are not
+    # hydrogens, this is an error. For excess hydrogens,
+    # the dictionary of protonation variants should contain a
+    # matching entry. However, it is currently very incomplete,
+    # meaning that many hydrogens cannot be verified, and do
+    # not have bond information.
+    extra_atoms = atom_ids.difference(comp_atom_ids)
+    if extra_atoms:
+        if not all(a.startswith('H') for a in extra_atoms):
+            names = [s[1]+'['+s[0]+']' for s in sites if s[1] in extra_atoms]
+            raise ValueError("Atom(s) %s not in component %s"
+                             % (', '.join(names), comp_id))
+        # Find the protonation variants for the current monomer
+        # that match the atom_ids.
+        vs = [v for v in variants.get(comp_id, [])
+              if set(components[v].atoms['atom_id']).issuperset(atom_ids)]
+        if len(vs) == 1:
+            # If there is exactly one matching variant, use it.
+            comp = components[vs[0]]
             comp_atom_ids = set(comp.atoms['atom_id'])
+            species = comp.three_letter_code
+            if hasattr(comp, 'bonds'):
+                bonds = tuple((mosaic_ids[x['atom_id_1']],
+                               mosaic_ids[x['atom_id_2']],
+                               bond_order(x.get('value_order')))
+                              for x in comp.bonds
+                              if x['atom_id_1'] in atom_ids
+                                 and x['atom_id_2'] in atom_ids)
+            else:
+                bonds = ()
+        else:
+            # Keep the main component and accept any additional hydrogens.
+            # Construct bonds for them based on a distance criterion.
+            p = dict((s[1], np.array(s[7:10])) for s in sites)
+            e = dict((s[1], s[6]) for s in sites)
+            rbsq = {'C': 0.115**2, 'N': 0.105**2, 'P': 0.16**2,
+                    'O': 0.105**2, 'S': 0.15**2}
+            for hydrogen_id in extra_atoms:
+                comp_atom_ids.add(hydrogen_id)
+                ph = p[hydrogen_id]
+                close = [atom_id
+                         for atom_id, position in p.items()
+                         if (not atom_id.startswith('H'))
+                            and np.sum((position-ph)**2) < rbsq[e[atom_id]]]
+                if len(close) > 1:
+                     name = [s[1]+'['+s[0]+']'
+                             for s in sites if s[1] == hydrogen_id][0]
+                     names = [s[1]+'['+s[0]+']'
+                              for s in sites if s[1] in close]
+                     raise ValueError("Cannot identify a unique bonding"
+                                      " partner for atom %s (close atoms"
+                                      " are %s)"
+                                      % (name, ', '.join(names)))
+                if len(close) > 0:
+                    bonds = bonds + ((mosaic_ids[close[0]],
+                                      mosaic_ids[hydrogen_id], 'single'),)
+
+    # Construct the MOSAIC fragment
     atom_sites = {}
     atom_ids = []
     atom_types = []
@@ -455,7 +594,7 @@ def make_monomer(sites, site_properties):
         atom_sites[atom_id] = atom_sites.get(atom_id, [])
         atom_sites[atom_id].append(unique_id)
         site_properties[unique_id] = (x, y, z, occupancy, u_iso)
-    atoms = tuple(Atom(atom_id, Element(atom_type.capitalize()),
+    atoms = tuple(Atom(mosaic_ids[atom_id], _make_atom_descriptor(atom_type),
                        len(atom_sites[atom_id]))
                   for atom_id, atom_type in zip(atom_ids, atom_types))
     unique_ids = tuple(sum((atom_sites[atom_id] for atom_id in atom_ids), []))
@@ -463,10 +602,21 @@ def make_monomer(sites, site_properties):
         label = comp.three_letter_code
     else:
         label = comp.three_letter_code + '_' + str(seq_id)
-        #if ins_code is not None:
-        #    label = label + '_' + ins_code
-    return Fragment(label, comp.id, (), atoms, bonds), unique_ids
+    return Fragment(label, species, (), atoms, bonds), unique_ids
 
+def _make_atom_descriptor(pdb_atom_type):
+    try:
+        # most atom_type values are chemical elements
+        return Element(pdb_atom_type.capitalize())
+    except ValueError:
+        match = re.match('([A-Z][A-Za-z]?)[0-9]*[+-]', pdb_atom_type)
+        if match:
+            # Ions with a charge indication - use the base element name
+            try:
+                return Element(match.group(1))
+            except ValueError:
+                pass
+        return Unknown(pdb_atom_type)
 
 def ss_bridges(structure, asym_id_1, asym_id_2):
     """
@@ -547,6 +697,25 @@ def symmetry_transformations(space_group):
         transformations.append((rot, tr))
     return transformations
 
+def chain_links(residues, atom1, atom2, bond_order):
+    def link_atoms(residue, atom):
+        if residue.atoms:
+            # standard monomer
+            atom_ref = residue.get(atom)
+            if atom_ref is None:
+                return []
+            else:
+                return [atom_ref]
+        else:
+            # heterogeneous monomer
+            fs = residue.fragments
+            return [f[atom] for f in fs if atom in f]
+    bonds = []
+    for r1, r2 in zip(residues[:-1], residues[1:]):
+        for a1 in link_atoms(r1, atom1):
+            for a2 in link_atoms(r2, atom2):
+                bonds.append((a1, a2, bond_order))
+    return tuple(bonds)
 
 def make_crystal(structure, model_number, universes):
     """
@@ -566,12 +735,6 @@ def make_crystal(structure, model_number, universes):
         entity = structure.entities[entity_id]
         entity_type = entity['type']
         if entity_type == 'polymer' and 'polymer_type' in entity:
-            # Sugars are classified as polymers in the PDB, but have
-            # no polymer type and a single residue with the
-            # non-numeric seq_id '.'. This looks like a bug in the PDB
-            # data, but since it's so frequent, we have to deal with
-            # it. We do so by treating polymers without a polymer_type
-            # as non-polymers.
             polymer_type = entity['polymer_type']
             if entity['polymer_nstd_linkage']:
                 raise ValueError("Non-standard linkage not yet implemented")
@@ -587,7 +750,12 @@ def make_crystal(structure, model_number, universes):
             # Start with empty residues
             residues = OrderedDict()
             for seq_id, comp_id in entity['sequence']:
-                f = Fragment(comp_id + '_' + str(seq_id), comp_id,
+                if isinstance(comp_id, tuple):
+                    # heterogeneous monomer
+                    species = '/'.join(comp_id)
+                else:
+                    species = comp_id
+                f = Fragment(species + '_' + str(seq_id), species,
                              (), (), ())
                 residues[seq_id] = f
                 monomer_fragments[(asym_id, seq_id)] = f
@@ -609,9 +777,7 @@ def make_crystal(structure, model_number, universes):
             residues = list(residues.values())
             # Add inter-monomer links
             if polymer_type in ['polypeptide(L)']:
-                peptide_bonds = tuple((r1.get('C'), r2.get('N'), 'single')
-                                      for r1, r2 in zip(residues[:-1],
-                                                        residues[1:]))
+                peptide_bonds = chain_links(residues, 'C', 'N', 'single')
                 ss_bonds = tuple((monomer_fragments[(asym_id, seq1)].get('SG'),
                                   monomer_fragments[(asym_id, seq2)].get('SG'),
                                   'single')
@@ -623,8 +789,7 @@ def make_crystal(structure, model_number, universes):
                                   'polydeoxyribonucleotide',
                                   'polydeoxyribonucleotide/'
                                       'polyribonucleotide hybrid']:
-                bonds = tuple((r1.get('O3\''), r2.get('P'), 'single')
-                              for r1, r2 in zip(residues[:-1], residues[1:]))
+                bonds = chain_links(residues, 'O3\'', 'P', 'single')
             elif polymer_type in ['polysaccharide(D)', 'polysaccharide(L)']:
                 # need to figure out how monosaccharides are linked
                 # example: 1AGA
@@ -645,19 +810,31 @@ def make_crystal(structure, model_number, universes):
             molecules[asym_id] = [Polymer(asym_id, name, residues, bonds,
                                           polymer_types.get(polymer_type,
                                                             polymer_type))]
+
+        elif entity_type == 'polymer' and 'polymer_type' not in entity:
+            # This is how sugars are described in the PDB, though
+            # this seems to be undocumented. Sugars are labelled as
+            # polymers, but have no polymer_type. They have one or
+            # two monomers but all atoms have '.' in the seq_id field.
+            f, u = make_multimer(asym_id, sites, site_properties,
+                                 structure.extra_bonds)
+            molecules[asym_id] = [f]
+            unique_ids[asym_id].extend(u)
+
         elif entity_type == 'water':
             ms = []
             for atom in sites:
                 if atom[6] == 'O':
                     ms.append([atom])
                 else:
-                    assert atom[6] == 'H'
+                    assert atom[6] in ['H', 'D']
                     ms[-1].append(atom)
             molecules[asym_id] = []
             for m in ms:
                 f, u = make_monomer(m, site_properties)
                 molecules[asym_id].append(f)
                 unique_ids[asym_id].extend(u)
+
         else:
             f, u = make_monomer(sites, site_properties)
             molecules[asym_id] = [f]
