@@ -1,9 +1,13 @@
 # -*- coding: utf-8 -*-
 
 # Tests for module mosaic.immutable_model
+
+#-----------------------------------------------------------------------------
+#       Copyright (C) 2013 The Mosaic Development Team
 #
-# Written by Konrad Hinsen
-#
+#  Distributed under the terms of the BSD License.  The full license is in
+#  the file LICENSE.txt, distributed as part of this software.
+#-----------------------------------------------------------------------------
 
 import unittest
 
@@ -122,6 +126,27 @@ class PeptideTest(unittest.TestCase):
         self.assertEqual(self._make_molecule(),
                          self._make_molecule())
 
+    def test_iterators(self):
+        mol = self._make_molecule()
+        mol_ref = M.FragmentRef('x', mol)
+        atoms = tuple(mol_ref.recursive_atom_iterator())
+        self.assertEqual(len(atoms), mol.number_of_atoms)
+        bonds = tuple(mol_ref.recursive_bond_iterator())
+        self.assertEqual(len(bonds), mol.number_of_bonds)
+        for a1, a2, order in bonds:
+            for a in a1, a2:
+                node = mol
+                for p in a.split('.'):
+                    node = node[p]
+                self.assertTrue(isinstance(node, M.Atom))
+        paths = tuple(mol_ref.recursive_atom_path_iterator())
+        self.assertEqual(len(paths), mol.number_of_atoms)
+        for ap in paths:
+            node = mol
+            for p in ap.split('.'):
+                node = node[p]
+            self.assertTrue(isinstance(node, M.Atom))
+
 class ErrorCheckingTest(unittest.TestCase):
 
     def test_atom_descriptor(self):
@@ -217,6 +242,10 @@ class ErrorCheckingTest(unittest.TestCase):
                           lambda: M.universe('infinite', [mol]))
         self.assertRaises(ValueError,
                           lambda: M.universe('infinite', [(10, mol)]))
+        self.assertRaises(ValueError,
+                          lambda: M.universe('infinite', [(mol, 'water', 10)],
+                                             [(IN.zeros((3,3), N.float64),
+                                               IN.zeros((3,), N.float64))]))
 
     def test_configuration(self):
         mol = make_water_fragment()
@@ -250,6 +279,30 @@ class ErrorCheckingTest(unittest.TestCase):
                                                   IN.zeros((30, 3), N.float32),
                                                   IN.zeros((3,), N.float32)))
         
+    def test_selection(self):
+        mol = make_water_fragment(nsites=2)
+        universe = M.universe('cube', [(mol, 'water', 5)])
+        # Index occurs twice
+        self.assertRaises(ValueError,
+                          lambda: M.AtomSelection(universe,
+                                                  IN.zeros((2,), N.uint16)))
+        # Atom index too large
+        self.assertRaises(ValueError,
+                          lambda: M.AtomSelection(universe,
+                                                  IN.array([20], N.uint16)))
+        # Template atom index too large
+        self.assertRaises(ValueError,
+                          lambda: M.TemplateAtomSelection(universe,
+                                                          IN.array([3], N.uint8)))
+        # Site index too large
+        self.assertRaises(ValueError,
+                          lambda: M.SiteSelection(universe,
+                                                  IN.array([40], N.uint16)))
+        # Template site index too large
+        self.assertRaises(ValueError,
+                          lambda: M.TemplateSiteSelection(universe,
+                                                          IN.array([8], N.uint8)))
+
 class UniverseTest(unittest.TestCase):
 
     def setUp(self):
@@ -355,6 +408,56 @@ class UniverseTest(unittest.TestCase):
         for s1, s2 in zip(labels, el.strings):
             self.assertEqual(s1, s2)
 
+    def test_bonds(self):
+        bonds = self.universe.bond_index_array()
+        self.assertEqual(len(bonds), self.universe.number_of_bonds)
+        self.assertTrue((bonds >= 0).all())
+        self.assertTrue((bonds < self.universe.number_of_atoms).all())
+        for i in range(10):
+            self.assertEqual(bonds[2*i, 0], 3*i)
+            self.assertEqual(bonds[2*i, 1], 3*i+2)
+            self.assertEqual(bonds[2*i+1, 0], 3*i+1)
+            self.assertEqual(bonds[2*i+1, 1], 3*i+2)
+
+    def test_index_mappings(self):
+        mol = self.universe.molecules[0][0]
+        s2a = mol.site_to_atom_index_mapping()
+        self.assertTrue((s2a == N.array([0, 0, 1, 1, 2, 2])).all())
+
+        s2a = self.universe.site_to_atom_index_mapping()
+        s2a_ref = N.repeat(N.arange(30), 2)
+        self.assertTrue((s2a == s2a_ref).all())
+
+        st2at = self.universe.template_site_to_template_atom_index_mapping()
+        st2at_ref = N.array([0, 0, 1, 1, 2, 2])
+        self.assertTrue((st2at == st2at_ref).all())
+
+        s2t = self.universe.site_to_template_index_mapping()
+        s2t_ref = N.resize(N.arange(mol.number_of_sites),
+                           (self.universe.number_of_sites,))
+        self.assertTrue((s2t == s2t_ref).all())
+
+        a2t = self.universe.atom_to_template_index_mapping()
+        a2t_ref = N.resize(N.arange(mol.number_of_atoms),
+                           (self.universe.number_of_atoms,))
+        self.assertTrue((a2t == a2t_ref).all())
+
+        a2s = self.universe.atom_to_site_index_mapping()
+        a2s_ref = 2*N.arange(self.universe.number_of_atoms)
+        self.assertTrue((a2s == a2s_ref).all())
+
+    def test_selections(self):
+        s = M.AtomSelection(self.universe, IN.array([0, 2], N.uint8))
+        self.assertEqual(s.number_of_atoms, 2)
+        self.assertEqual(s.number_of_sites, 4)
+        s = M.TemplateAtomSelection(self.universe, IN.array([1], N.uint8))
+        self.assertEqual(s.number_of_atoms, 10)
+        self.assertEqual(s.number_of_sites, 20)
+        s = M.SiteSelection(self.universe, [2])
+        self.assertEqual(s.number_of_sites, 1)
+        s = M.TemplateSiteSelection(self.universe, [0])
+        self.assertEqual(s.number_of_sites, 10)
+
 class PBCTest(unittest.TestCase):
 
     def setUp(self):
@@ -368,6 +471,7 @@ class PBCTest(unittest.TestCase):
                                IN.zeros((0, 3), N.float32),
                                None)
         self.assertEqual(conf.lattice_vectors(), ())
+        self.assertEqual(conf.cell_volume(), None)
         conf = M.Configuration(self.cube,
                                IN.zeros((0, 3), N.float32),
                                IN.array(1., N.float32))
@@ -375,6 +479,7 @@ class PBCTest(unittest.TestCase):
         self.assertTrue((lv[0] == N.array([1., 0., 0.], N.float32)).all())
         self.assertTrue((lv[1] == N.array([0., 1., 0.], N.float32)).all())
         self.assertTrue((lv[2] == N.array([0., 0., 1.], N.float32)).all())
+        self.assertEqual(conf.cell_volume(), 1.)
         conf = M.Configuration(self.cuboid,
                                IN.zeros((0, 3), N.float32),
                                IN.array([1., 2., 4.], N.float32))
@@ -382,6 +487,7 @@ class PBCTest(unittest.TestCase):
         self.assertTrue((lv[0] == N.array([1., 0., 0.], N.float32)).all())
         self.assertTrue((lv[1] == N.array([0., 2., 0.], N.float32)).all())
         self.assertTrue((lv[2] == N.array([0., 0., 4.], N.float32)).all())
+        self.assertEqual(conf.cell_volume(), 8.)
         conf = M.Configuration(self.parallelepiped,
                                IN.zeros((0, 3), N.float32),
                                IN.array([[1., 2., 4.],
@@ -391,9 +497,7 @@ class PBCTest(unittest.TestCase):
         self.assertTrue((lv[0] == N.array([1., 2., 4.], N.float32)).all())
         self.assertTrue((lv[1] == N.array([8., 4., 2.], N.float32)).all())
         self.assertTrue((lv[2] == N.array([16., 4., 8.], N.float32)).all())
-
-
-PBCTest('test_lattice_vectors').debug()
+        self.assertAlmostEqual(conf.cell_volume(), 168.)
 
 def suite():
     loader = unittest.TestLoader()
