@@ -1,9 +1,13 @@
 # -*- coding: utf-8 -*-
 
 # Tests for module mosaic.mutable_model
+
+#-----------------------------------------------------------------------------
+#       Copyright (C) 2013 The Mosaic Development Team
 #
-# Written by Konrad Hinsen
-#
+#  Distributed under the terms of the BSD License.  The full license is in
+#  the file LICENSE.txt, distributed as part of this software.
+#-----------------------------------------------------------------------------
 
 import copy
 import unittest
@@ -161,6 +165,26 @@ class PeptideTest(unittest.TestCase):
         self.assertFalse(self._make_molecule('di_ala_1')
                         .is_equivalent(self._make_molecule('di_ala_2')))
 
+    def test_iterators(self):
+        mol = self._make_molecule('di_ala')
+        atoms = tuple(mol.recursive_atom_iterator())
+        self.assertEqual(len(atoms), mol.number_of_atoms)
+        bonds = tuple(mol.recursive_bond_iterator())
+        self.assertEqual(len(bonds), mol.number_of_bonds)
+        for a1, a2, order in bonds:
+            for a in a1, a2:
+                node = mol
+                for p in a.split('.'):
+                    node = node[p]
+                self.assertTrue(isinstance(node, M.Atom))
+        paths = tuple(mol.recursive_atom_path_iterator())
+        self.assertEqual(len(paths), mol.number_of_atoms)
+        for ap in paths:
+            node = mol
+            for p in ap.split('.'):
+                node = node[p]
+            self.assertTrue(isinstance(node, M.Atom))
+
 class ErrorCheckingTest(unittest.TestCase):
 
     def test_atom_descriptor(self):
@@ -270,6 +294,10 @@ class ErrorCheckingTest(unittest.TestCase):
                           lambda: M.Universe('infinite', [mol]))
         self.assertRaises(ValueError,
                           lambda: M.Universe('infinite', [(10, mol)]))
+        self.assertRaises(ValueError,
+                          lambda: M.Universe('infinite', [(mol, 10)],
+                                             [(N.zeros((3,3), N.float64),
+                                               N.zeros((3,), N.float64))]))
 
     def test_configuration(self):
         mol = make_water_fragment()
@@ -309,6 +337,30 @@ class ErrorCheckingTest(unittest.TestCase):
                                                   N.float32(10.),
                                                   N.float64))
         
+    def test_selection(self):
+        mol = make_water_fragment(nsites=2)
+        universe = M.Universe('cube', [(mol, 5)])
+        # Index occurs twice
+        self.assertRaises(ValueError,
+                          lambda: M.AtomSelection(universe,
+                                                  N.zeros((2,), N.uint16)))
+        # Atom index too large
+        self.assertRaises(ValueError,
+                          lambda: M.AtomSelection(universe,
+                                                  N.array([20], N.uint16)))
+        # Template atom index too large
+        self.assertRaises(ValueError,
+                          lambda: M.TemplateAtomSelection(universe,
+                                                          N.array([3], N.uint8)))
+        # Site index too large
+        self.assertRaises(ValueError,
+                          lambda: M.SiteSelection(universe,
+                                                  N.array([40], N.uint16)))
+        # Template site index too large
+        self.assertRaises(ValueError,
+                          lambda: M.TemplateSiteSelection(universe,
+                                                          N.array([8], N.uint8)))
+
 class UniverseTest(unittest.TestCase):
 
     def setUp(self):
@@ -410,6 +462,56 @@ class UniverseTest(unittest.TestCase):
         for s1, s2 in zip(labels, el.strings):
             self.assertEqual(s1, s2)
 
+    def test_bonds(self):
+        bonds = self.universe.bond_index_array()
+        self.assertEqual(len(bonds), self.universe.number_of_bonds)
+        self.assertTrue((bonds >= 0).all())
+        self.assertTrue((bonds < self.universe.number_of_atoms).all())
+        for i in range(10):
+            self.assertEqual(bonds[2*i, 0], 3*i)
+            self.assertEqual(bonds[2*i, 1], 3*i+2)
+            self.assertEqual(bonds[2*i+1, 0], 3*i+1)
+            self.assertEqual(bonds[2*i+1, 1], 3*i+2)
+
+    def test_index_mappings(self):
+        mol = self.universe.molecules[0][0]
+        s2a = mol.site_to_atom_index_mapping()
+        self.assertTrue((s2a == N.array([0, 0, 1, 1, 2, 2])).all())
+
+        s2a = self.universe.site_to_atom_index_mapping()
+        s2a_ref = N.repeat(N.arange(30), 2)
+        self.assertTrue((s2a == s2a_ref).all())
+
+        st2at = self.universe.template_site_to_template_atom_index_mapping()
+        st2at_ref = N.array([0, 0, 1, 1, 2, 2])
+        self.assertTrue((st2at == st2at_ref).all())
+
+        s2t = self.universe.site_to_template_index_mapping()
+        s2t_ref = N.resize(N.arange(mol.number_of_sites),
+                           (self.universe.number_of_sites,))
+        self.assertTrue((s2t == s2t_ref).all())
+
+        a2t = self.universe.atom_to_template_index_mapping()
+        a2t_ref = N.resize(N.arange(mol.number_of_atoms),
+                           (self.universe.number_of_atoms,))
+        self.assertTrue((a2t == a2t_ref).all())
+
+        a2s = self.universe.atom_to_site_index_mapping()
+        a2s_ref = 2*N.arange(self.universe.number_of_atoms)
+        self.assertTrue((a2s == a2s_ref).all())
+
+    def test_selections(self):
+        s = M.AtomSelection(self.universe, N.array([0, 2], N.uint8))
+        self.assertEqual(s.number_of_atoms, 2)
+        self.assertEqual(s.number_of_sites, 4)
+        s = M.TemplateAtomSelection(self.universe, N.array([1], N.uint8))
+        self.assertEqual(s.number_of_atoms, 10)
+        self.assertEqual(s.number_of_sites, 20)
+        s = M.SiteSelection(self.universe, [2])
+        self.assertEqual(s.number_of_sites, 1)
+        s = M.TemplateSiteSelection(self.universe, [0])
+        self.assertEqual(s.number_of_sites, 10)
+
 class PBCTest(unittest.TestCase):
 
     def setUp(self):
@@ -422,6 +524,7 @@ class PBCTest(unittest.TestCase):
         conf = M.Configuration(self.infinite,
                                N.zeros((0, 3), N.float32))
         self.assertEqual(conf.lattice_vectors(), ())
+        self.assertEqual(conf.cell_volume(), None)
         conf = M.Configuration(self.cube,
                                N.zeros((0, 3), N.float32),
                                N.array(1., N.float32))
@@ -429,6 +532,7 @@ class PBCTest(unittest.TestCase):
         self.assertTrue((lv[0] == N.array([1., 0., 0.])).all())
         self.assertTrue((lv[1] == N.array([0., 1., 0.])).all())
         self.assertTrue((lv[2] == N.array([0., 0., 1.])).all())
+        self.assertEqual(conf.cell_volume(), 1.)
         conf = M.Configuration(self.cuboid,
                                N.zeros((0, 3), N.float32),
                                N.array([1., 2., 4.], N.float32))
@@ -436,6 +540,7 @@ class PBCTest(unittest.TestCase):
         self.assertTrue((lv[0] == N.array([1., 0., 0.])).all())
         self.assertTrue((lv[1] == N.array([0., 2., 0.])).all())
         self.assertTrue((lv[2] == N.array([0., 0., 4.])).all())
+        self.assertEqual(conf.cell_volume(), 8.)
         conf = M.Configuration(self.parallelepiped,
                                N.zeros((0, 3), N.float32),
                                N.array([[1., 2., 4.],
@@ -445,6 +550,7 @@ class PBCTest(unittest.TestCase):
         self.assertTrue((lv[0] == N.array([1., 2., 4.])).all())
         self.assertTrue((lv[1] == N.array([8., 4., 2.])).all())
         self.assertTrue((lv[2] == N.array([16., 4., 8.])).all())
+        self.assertAlmostEqual(conf.cell_volume(), 168.)
         
 def suite():
     loader = unittest.TestLoader()
